#!/usr/bin/perl -w
use Getopt::Long qw(GetOptions);
use IO::Socket::INET;

# play_ais
# Program designed to read timestamped AIS data created with the timestamp_data
# program in order to recreate a real-time replay of accumulated data

# Socket code inspired by:
# https://www.thegeekstuff.com/2010/07/perl-tcp-udp-socket-programming/

# timestamp itself assumed to be from the time() function and are, therefore, in seconds.
# Thus, any timestamp format will do as long as the relative delta is seconds. This program
# does NOT send any AIS timestamp information.

# Set parameters

$version = "2.3.2";
$build_date = "05/24/2022";

# Set default values for all switches so that non-supplied values don't cause
# "uninitialized value" errors

$host = "127.0.0.1"; # localhost
$port = 0; 
$protocol = "tcp";
$verbose = "";
$rate = 1;
$help = "";
$test_mode = 0;

$time_field = 6;
$ais_field = 7;
$sep = "|";

# Parse the command line to get message parameters

GetOptions
  (
  "help" => \$help,
  "test" => \$test_mode,
  "filename=s" => \$filename,
  "iphost=s" => \$host,
  "port=i" => \$port,
  "protocol=s" => \$protocol,
  "rate=s" => \$rate,
  "separator=s" => \$sep,
  "ais=i" => \$ais_field,
  "time=i" => \$time_field,
  "verbose" => \$verbose,
  )
  or die "Usage: $0 --help for command line assistance\n";

# Check values...

print "\nAIS Play (Version $version, Build date: $build_date)\n";

# If the -h switch is set, just print the help file.

if ($help)
  {
  help ();
  exit;
  }

# If the test_mode switch is set, also set verbose

if ($test_mode == 1) { $verbose = 1; }

# Ensure that the separator character is not interpreted as a special regex character

$sep = quotemeta ($sep);

# If NOT in test mode, set up output parameters and make TCP/UDP connection

if ($test_mode == 0)
  {
  if ($port <= 0 || $port > 65535)
    {
    print "\nInvalid port number $port. Program terminating...\n\n";
    exit;
    }

  if (lc($protocol) ne "tcp" && lc($protocol) ne "udp")
    {
    print "\nInvalid protocol $protocol; using TCP\n";
    $protocol = "tcp";
    }

  $socket = new IO::Socket::INET (
  PeerHost => $host,
  PeerPort => $port,
  Proto => $protocol,
  ) or die "ERROR in Socket Creation : $!\n";

  print "Connection success to $host:$port/$protocol.\n";
  }

# Get the data source...

open (INFILE, '<:encoding(UTF-8)', $filename)
  or die "Could not open file '$filename'\n";

print "Reading file $filename, using '$sep' as a field separator.\n";
print "Timestamps are in field $time_field and AIS data in field $ais_field.\n\n";

$num = 0;
$old_time = 0;

while (not (eof INFILE))
  {
  $line = <INFILE>;
  chomp $line;
  $num++;

# Split fields in $line. Find current time and the AIS sentence.

  @fields = split ($sep, $line);
  $current_time = $fields [$time_field];
  $sentence = $fields [$ais_field];

# Determine how long to wait...

  if ($current_time == 0 || $current_time <= $old_time)
      { $wait = 0; }
    else
      { $wait = ($current_time - $old_time) / $rate; }
  sleep $wait;

  if ($test_mode == 0) { print $socket "$sentence\n"; }
  if ($verbose) { print "Record $num -- Timestamp = $current_time; waited $wait second(s) to send...\n $sentence\n"; }

  $old_time = $current_time;
  }

print "\n$num records sent\n\n";

if ($test_mode == 0) { $socket->close(); }
close INFILE;

# ********************************************************
# help
# ********************************************************

# Display the help information for the program

sub help
  {
print<< "EOT";

play_ais reads data from a file and writes the data to a TCP/UDP socket.

There are eight parameters:

--filename   Input file name
--iphost     A host name or IP address. NO ERROR CHECKING IS PERFORMED ON THIS PARAMETER
             [Default = 127.0.0.1 (localhost)]
--port       TCP/UDP port number. Must be in range 1-65535
--protocol   TCP or UDP [Default = TCP]
--rate       Rate of playback to real-time [Default = 1]
             If rate>1, playback is faster than real-time; if rate <1, playback is slower
--verbose    If set, output goes to STDOUT as well as to a TCP/UDP socket
--separator  Specify the field separator [Default = |]
             NOTE: Select a separator unlikely to appear in an AIS message, such as "|" or "-"
--ais        The field containing the AIS sentence [Default = 7]
--time       The field containing the seconds timestamp [Default = 6]
             NOTE: FIELD NUMBERS START AT 0 FOR --ais AND --time. These default values are
             compatible with output from timestamp_data.pl
--test       Puts the program into test (and verbose) mode. Runs normally but only displays
             output without opening a TCP/UDP socket

EXAMPLE: perl play_ais.pl --filename=test.txt --port=26 --separator=\| --verbose
         ./play_ais.pl -f=test.txt -ip=out.example.net -pr=tcp -po=12345 -s=- -a=1 -t=0 -v

Use -h or --help to obtain this help message.

EOT
  return;
  }

